/* %Z%%M% %I% %E% 1989 by Robin T. Miller. */

/****************************************************************************
 *									    *
 *			  COPYRIGHT (c) 1989 - 1995			    *
 *			   This Software Provided			    *
 *				     By					    *
 *			   MILLER W J & ASSOCIATES			    *
 *			       14 Aladdin Road 				    *
 *			      Windham, NH 03087				    *
 *			       (603) 890-8723				    *
 *									    *
 * Permission to use, copy, modify, distribute and sell this software and   *
 * its documentation for any purpose and without fee is hereby granted,	    *
 * provided that the above copyright notice appear in all copies and that   *
 * both that copyright notice and this permission notice appear in the	    *
 * supporting documentation, and that the name of the author not be used    *
 * in advertising or publicity pertaining to distribution of the software   *
 * without specific, written prior permission.				    *
 *									    *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, 	    *
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN	    *
 * NO EVENT SHALL HE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL   *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR    *
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  *
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF   *
 * THIS SOFTWARE.							    *
 *									    *
 ****************************************************************************/
/*
 * Modification History:
 *
 * November 22, 1995 by Robin Miller.
 *	Initial port to QNX 4.21 Operating System.
 */

/************************************************************************
 *									*
 * Program:	tape_dup - Tape Duplication Program.			*
 *									*
 * Date:	September 22, 1989					*
 *									*
 * Author:	Robin T. Miller						*
 *									*
 ************************************************************************/

#include "tape_dup.h"
#include <malloc.h>
#include <signal.h>

int	compare_flag = FALSE;		/* Compare tape with disk files	*/
int	debug_flag = FALSE;		/* Enable debug output flag.	*/
int	verbose_flag = TRUE;		/* Verbose messages output.	*/
int	verify_flag = FALSE;		/* Verify the read/write data.	*/
char	*name;				/* The program name.		*/
iobuf_t	compare_buffer;			/* Buffer for data comparisons.	*/
iobuf_t	data_buffer;			/* Pointer to malloc'ed buffer.	*/
char	*directory_name = ".";		/* Directory name of data files.*/
char	*input_file;			/* Pointer to input file name.	*/
char	*output_file;			/* Pointer to output file name.	*/
char	*operation_type = "unknown";	/* Type of operation performed.	*/
int	block_size = DEFAULT_SIZE;	/* Default block size to use.	*/
int	file_count = BIG;		/* Number of files to process.	*/
int	file_number = 1;		/* The current file number.	*/
long	file_size;			/* The current file size.	*/
long	record_count;			/* The current record count.	*/
long	starting_record;		/* The starting record number.	*/
int	total_files;			/* The total number of files.	*/
long	total_size;			/* The total file size (bytes).	*/
int	skip_count;			/* # of input files to skip.	*/
int	last_read;			/* The last read count.		*/
char	*string;			/* Current string pointer.	*/
int	ifd = -1, ofd = -1;		/* Input/output file descrip.	*/
FILE	*fp, *fpi;			/* File pointers (disk files).	*/
clock_t after_time, before_time;	/* Per pass elapsed time.	*/
struct timeb stime, btime, etime;	/* For elapsed times.		*/

/************************************************************************
 *									*
 * main()	Tape duplication program.				*
 *									*
 * Description:	This program is used for duplicating tapes.  The tape	*
 *		read function reads tape files onto the disk, and the	*
 *		tape write function writes these files back to tape.	*
 *		When reading files from tape, an information file is	*
 *		created which is used to ensure data integtriy when	*
 *		writing the file back to tape.				*
 *									*
 ************************************************************************/
int
main (argc, argv)
int argc;
char **argv;
{
	char *tmp;
	int c, status;

	tmp = rindex (argv[0], '/');
	name = tmp ? &(tmp[1]) : argv[0];

	for (c = 1; c < argc; c++) {
	    string = argv[c];
	    if (match ("bs=")) {
		block_size = number(BIG);
		if (block_size <= 0) {
		    fprintf (stderr,"%s: block_size must be postive.\n", name);
		    exit (2);
		}
		continue;
	    }
	    if (match ("enable=")) {
	    eloop:
		if (match(","))
		    goto eloop;
		if (*string == '\0')
		    continue;
#ifdef notdef
		if (match("compare")) {
		    compare_flag = TRUE;
		    operation_type = "compare";
		    goto eloop;
		}
#endif /* notdef */
		if (match("debug")) {
		    debug_flag = TRUE;
		    goto eloop;
		}
		if (match("verbose")) {
		    verbose_flag = TRUE;
		    goto eloop;
		}
		if (match("verify")) {
		    operation_type = "verify";
		    verify_flag = TRUE;
		    if (!input_file && !output_file) {
			output_file = "/dev/null";
		    }
		    goto eloop;
		}
		fprintf (stderr, "%s: Invalid enable keyword: %s\n",
							name, string);
		exit (1);
	    }
	    if (match ("disable=")) {
	    dloop:
		if (match(","))
		    goto dloop;
		if (*string == '\0')
		    continue;
#ifdef notdef
		if (match("compare")) {
		    compare_flag = FALSE;
		    goto dloop;
		}
#endif /* notdef */
		if (match("verbose")) {
		    verbose_flag = FALSE;
		    goto dloop;
		}
		if (match("verify")) {
		    verify_flag = FALSE;
		    goto dloop;
		}
		fprintf (stderr, "%s: Invalid disable keyword: %s\n",
							name, string);
		exit (1);
	    }
	    if (match ("directory=")) {
		directory_name = string;
		continue;
	    }
	    if (match ("files=")) {
		file_count = number(BIG);
		continue;
	    }
	    if (match ("if=")) {
		input_file = string;
		operation_type = "read";
		continue;
	    }
	    if (match ("of=")) {
		output_file = string;
		operation_type = "write";
		continue;
	    }
	    if (match ("skip=")) {
		skip_count = number(BIG);
		continue;
	    }
	    usage();
	}

	if ( (!input_file && !output_file) || (input_file && output_file) ) {
	    fprintf (stderr,
		"%s: You must specify only 1 input file or output file.\n",
								name);
	    exit (1);
	}
	if (input_file) {
	    if ((ifd = open (input_file, O_RDONLY, 0666)) < 0) {
		fprintf (stderr,"Can't open %s\n", input_file);
		perror ("open");
		exit (1);
	    }
	} else {
	    if (verify_flag == FALSE) {
		if ((ofd = open (output_file, O_WRONLY, 0666)) < 0) {
		    fprintf (stderr,"Can't open %s\n", output_file);
		    perror ("open");
		    exit (3);
		}
	    }
	}

	/*
	 * Allocate data buffers.
	 */
	if ( (data_buffer = (iobuf_t) malloc (block_size)) == (iobuf_t) 0) {
		fprintf (stderr, "%s: malloc of data buffer failed.\n", name);
		exit (2);
	}
#ifdef notdef
	if (compare_flag == TRUE) {
	    if ( (compare_buffer = (iobuf_t) malloc (block_size)) == (iobuf_t) 0) {
		fprintf (stderr, "%s: malloc of compare buffer failed.\n", name);
		exit (2);
	    }
	}
#endif /* notdef */

	/*
	 * Catch a couple signals to do elegant cleanup.
	 */
	(void) signal (SIGHUP, terminate);
	(void) signal (SIGINT, terminate);
	(void) signal (SIGTERM, terminate);

	/*
	 * Dispatch to appropriate function.
	 */
#if defined(TIME_TAPE)
	if (verbose_flag) {
		ftime (&stime);
	}
#endif /* defined(TIME_TAPE) */
	if (input_file) {
		status = tape_read();
	} else {
		status = tape_write();
	}
	if (verbose_flag && (status != SUCCESS)) {
	    fprintf (stderr, "%s: Operation Aborted\n", name);
	}
	terminate (status);
}

#if defined(TIME_TAPE)
/************************************************************************
 *									*
 * report_time() - Report the elapsed time.				*
 *									*
 ************************************************************************/
void
report_time(void)
{
	ftime (&etime);
	printf (", elapsed = ");
	after_time = HZ * (etime.time - btime.time);
	after_time += ((HZ * ((etime.millitm-btime.millitm)))/1000);
	print_time (after_time);
	printf ("\n");
	fflush (stdout);
}


/************************************************************************
 *									*
 * print_time() - Convert binary ASCII time and print it.		*
 *									*
 ************************************************************************/
void
print_time (clock_t time)
{
	clock_t hr, min, sec, frac;

	frac = time % HZ;
	frac = (frac * 100) / HZ;
	time /= HZ;
	sec = time % 60; time /= 60;
	min = time % 60;
	if (hr = time / 60) {
	    printf ("%dh", hr);
	}
	printf ("%02dm", min);
	printf ("%02d.", sec);
	printf ("%02ds", frac);
}
#else /* !defined(TIME_TAPE) */
void
report_time()
{
	printf (".\n");
}
#endif /* defined(TIME_TAPE) */

void
terminate (int code)
{
#if defined(TIME_TAPE)
	struct timeb *stm;
#endif /* defined(TIME_TAPE) */

	if (verbose_flag) {
	    if (file_size) {			/* Partial file processed. */
		total_size += file_size;
	    }
	    printf ("A total of %d files consisting of %d bytes were processed.\n",
						file_number, total_size);
#if defined(TIME_TAPE)
	    ftime (&etime);
	    stm = &stime;
	    printf ("The total time to %s %.2f Kbytes (%.2f Mbytes) was ",
				operation_type,
				((float)total_size / (float)KBYTE),
				((float)total_size / (float)MBYTE));
	    after_time = HZ * (etime.time - stm->time);
	    after_time += ((HZ * ((long)(etime.millitm-stm->millitm)))/1000);
	    print_time (after_time);
	    printf (".\n");
#endif /* defined(TIME_TAPE) */
	}
	exit (code);
}

int
match (char *s)
{
	register char *cs;

	cs = string;
	while (*cs++ == *s)
		if (*s++ == '\0')
			goto true;
	if (*s != '\0')
		return(0);

true:
	cs--;
	string = cs;
	return (1);
}

int
number (int big)
{
	register char *cs;
	long n;

	cs = string;
	n = 0;
	while (*cs >= '0' && *cs <= '9') {
		n = n*10 + *cs++ - '0';
	}
	for (;;) {
	    switch (*cs++) {

		case 'k':
		case 'K':			/* Kilobytes */
			n *= KBYTE;
			continue;

		case 'm':
		case 'M':			/* Megabytes */
			n *= MBYTE;
			continue;

		case 'w':
		case 'W':			/* Word count. */
			n *= sizeof(int);
			continue;

		case 'b':
		case 'B':			/* Block count. */
			n *= BLOCK;
			continue;

		case '*':
		case 'x':
		case 'X':
			string = cs;
			n *= number(BIG);

		case '\0':
			if (n>=big || n<0) {
			    fprintf (stderr,
				"%s: argument %D out of range\n", name, n);
			    exit (2);
			}
			return(n);

		default:
			fprintf (stderr, "%s: invalid character in number: %c\n",
		    		name, cs[-1]);
			exit (2);
	    }
	}
}

/*
 * usage()	Display valid options for tape_copy program.
 */
void
usage(void)
{
    fprintf (stderr, "Usage: %s -options ...\n", name);
    fprintf (stderr, "\n    Where -options are:\n");
    fprintf (stderr, "\tif=filename    The input file to read.\n");
    fprintf (stderr, "\tof=filename    The output file to read.\n");
    fprintf (stderr, "\tbs=value       The block size to read/write.\n");
    fprintf (stderr, "\tdirectory=path The directory name of data files.\n");
    fprintf (stderr, "\tfiles=value    The number of files to process.\n");
    fprintf (stderr, "\tenable=flag    Enable one or more of the flags below.\n");
    fprintf (stderr, "\tdisable=flag   Disable one or more of the flags below.\n");
    fprintf (stderr, "\n    Flags to enable/disable:\n");
#ifdef notdef
    fprintf (stderr, "\tcompare        Compare tape to disk. (Default: disabled)\n");
#endif /* notdef */
    fprintf (stderr, "\tdebug          Debug output.         (Default: disabled)\n");
    fprintf (stderr, "\tverbose        Verbose output.       (Default: enabled)\n");
    fprintf (stderr, "\tverify         Verify data.          (Default: disabled)\n");
    exit (1);
}
