/* %Z%%M% %I% %E% 1989 by Robin T. Miller. */

/****************************************************************************
 *									    *
 *			  COPYRIGHT (c) 1989 - 1995			    *
 *			   This Software Provided			    *
 *				     By					    *
 *			   MILLER W J & ASSOCIATES			    *
 *			       14 Aladdin Road 				    *
 *			      Windham, NH 03087				    *
 *			       (603) 890-8723				    *
 *									    *
 * Permission to use, copy, modify, distribute and sell this software and   *
 * its documentation for any purpose and without fee is hereby granted,	    *
 * provided that the above copyright notice appear in all copies and that   *
 * both that copyright notice and this permission notice appear in the	    *
 * supporting documentation, and that the name of the author not be used    *
 * in advertising or publicity pertaining to distribution of the software   *
 * without specific, written prior permission.				    *
 *									    *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, 	    *
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN	    *
 * NO EVENT SHALL HE BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL   *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR    *
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  *
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF   *
 * THIS SOFTWARE.							    *
 *									    *
 ****************************************************************************/
/*
 * Modification History:
 *
 * November 22, 1995 by Robin Miller.
 *	Initial port to QNX 4.21 Operating System.
 */
#include "tape_dup.h"

/************************************************************************
 *									*
 * tape_read() - Read file from tape.					*
 *									*
 * Description:	This function is used to read files from a tape.  There	*
 *		are two disk files created for each tape file read. The	*
 *		first file contains the actual data, and the second	*
 *		file contains information about each record within the	*
 *		data file (a record is defined by the block size used).	*
 *		The information file contains data which is used when	*
 *		writing the files back to tape to ensure data intrigity.*
 *									*
 * Outputs:	Returns SUCCESS/FAILURE.				*
 *									*
 ************************************************************************/
int
tape_read()
{
	register u_short crc = 0;
	ssize_t read_count, write_count;
	int status = SUCCESS;
	int end_of_file = TRUE;
	int end_of_tape = FALSE;

	last_read = -2;				/* Flag no read done yet. */
	/*
	 * Loop reading all or specified files to/from the tape.
	 */
	do {
	    /*
	     * Read a file and write it to disk.
	     */	    
	    record_count++;
	    if ((read_count = read (ifd, data_buffer, block_size)) > 0) {
		/*
		 * If start of a new file, open new disk files.
		 */
		if (end_of_file == TRUE) {
		    end_of_file = FALSE;
		    if ((status = tr_open_files()) != SUCCESS) {
			return (status);
		    }
		}
		/*
		 * Write the data to the disk file.
		 */
		if (verify_flag == FALSE) {
		    if ((write_count = fwrite (data_buffer, sizeof(char), read_count, fp)) != read_count) {
			fprintf (stderr, "%s: fwrite (%d) != read (%d)\n",
					name, write_count, read_count);
			return (EIO);
		    }
		}
		/*
		 * Calculate the crc and update the information file.
		 */
		crc = 0;
		file_size += read_count;
		if (verify_flag == FALSE) {
		    crc = updcrc (crc, data_buffer, read_count);
		    if ((fprintf (fpi, "%d %d %x\n",
				record_count, read_count, crc)) == EOF) {
			fprintf (stderr,
			    "%s: Error writing info file record.\n", name);
			return (EIO);
		    }
		}
		if (read_count != last_read) {
		    if ( (last_read > 0) && debug_flag) {
			if (record_count - starting_record > 1) {
			    printf ("file %d: records %ld to %ld: size %d\n",
				file_number, starting_record, record_count-1, last_read);
			} else {
			    printf ("file %d: record %ld: size %d\n",
				file_number, starting_record, last_read);
			}
		    }
		    last_read = read_count;
		    starting_record = record_count;
		}
#if defined(_QNX_SOURCE)
		/*
		 * On QNX, if a tape file mark is encountered while reading
		 * a record, a short read is returned, but the end of file
		 * is *not* saved for the next read request, unlike UNIX's.
		 * [ NOTE: This isn't a big deal at this time, since QNX
		 *   does not currently support variable length records. ]
		 */
		if (read_count < block_size) {
		    goto end_of_file;
		}
#endif /* defined(_QNX_SOURCE) */
	    } else {			/* Error or end of file. */
end_of_file:
		if ( (last_read <= 0) && (last_read != -2) ) {
		    if (debug_flag) printf ("End of tape detected.\n");
		    if (verify_flag == FALSE) {
			(void) fclose (fp);
			(void) fclose (fpi);
		    }
		    end_of_tape = TRUE;
		    continue;
		}
		if ( (last_read > 0) && debug_flag) {
		    if (record_count - starting_record > 1) {
			printf ("file %d: records %ld to %ld: size %d\n",
					file_number, starting_record,
					record_count-1, last_read);
		    } else {
			printf ("file %d: record %ld: size %d\n",
				file_number, starting_record, last_read);
		    }
		}
		if (--record_count == 0) {
		    file_number--;
		} else if (verbose_flag || debug_flag) {
		    printf ("File %d: Processed %d records containing %d bytes",
				file_number, record_count, file_size);
		    report_time();
		}
		/*
		 * Format and output the file totals.
		 */
		if ( (verify_flag == FALSE) && (fpi != NULL) ) {
		    if ((fprintf (fpi,  "%d %d %x\n", -1, file_size, 0)) == EOF) {
		    	fprintf (stderr,
			    "%s: Error writing totals to info file.\n", name);
			return (EIO);
		    }
		}
		file_number++;
		record_count = 0;
		starting_record = 0;
		total_size += file_size;
		file_size = 0;
		end_of_file = TRUE;
		if (verify_flag == FALSE) {
		    if (fp != NULL) {
			(void) fclose (fp);
			fp = NULL;
		    }
		    if (fpi != NULL) {
			(void) fclose (fpi);
			fpi = NULL;
		    }
		}
/* Bug */
#ifdef notdef
		/*
		 * Must close and reopen tape to reset EOF flag.
		 */
		(void) close (ifd);
		if ((ifd = open (input_file, O_RDONLY, 0666)) < 0) {
		    fprintf (stderr,"Can't open %s\n", input_file);
		    return (1);
		}
#endif
/* Bug */
		if (file_number > file_count) {
		    break;
		}
		last_read = read_count;
	    }
	} while (end_of_tape == FALSE);

	(void) close (ifd);
	file_number--;
	/*
	 * Create file with the total files count in it.
	 */
	if (verify_flag == FALSE) {
	    char tbuff[DIR_SIZE];
	    char *bufp = tbuff;

	    (void) sprintf (bufp, "%s/%s", directory_name, FILE_TOTALS);
	    if ( (fp = fopen (bufp, "w")) == NULL) {
		fprintf (stderr, "%s: Unable to open totals file %s\n",
							name, bufp);
		return (FAILURE);
	    }
	    if ((fprintf (fp, "%d\n", total_files)) == EOF) {
		fprintf (stderr, "%s: Error writing to totals file.\n", name);
		fprintf (stderr, "%s: Totals file NOT written.\n", name);
		status = EIO;
	    }
	    (void) fclose (fp);
	}
	return (status);
}

/************************************************************************
 *									*
 * tr_open_files() - Open disk file to write tape data to.		*
 *									*
 * Description:	This function is used to open the disk files used when	*
 *		writing the next tape file to disk.  Both a data file	*
 *		and an information file will be created.		*
 *									*
 * Implicit Inputs:							*
 *		file_number - Used to construct the file names to open.	*
 *									*
 * Outputs:	Returns SUCCESS/FAILURE.				*
 *									*
 ************************************************************************/
int
tr_open_files()
{
	char tbuff[DIR_SIZE];
	char *bufp = tbuff;

	total_files++;
	if (verify_flag == FALSE) {
	    (void) sprintf (bufp, "%s/file_%d", directory_name, file_number);
	    if ( (fp = fopen (bufp, "w")) == NULL) {
		fprintf (stderr, "%s: Unable to open output file %s\n",
							name, bufp);
		return (FAILURE);
	    }
	    strcat (bufp, ".info");
	    if ( (fpi = fopen (bufp, "w")) == NULL) {
		fprintf (stderr, "%s: Unable to open info file %s\n",
							name, bufp);
		return (FAILURE);
	    }
	}
	if (debug_flag) {
	    printf ("Reading file number %d...\n", file_number);
	}
#if defined(TIME_TAPE)
	ftime (&btime);
#endif /* defined(TIME_TAPE) */
	return (SUCCESS);
}
